AMQP_MANAGER_EVENTS_X = "exchangeManagerEvents"
AMQP_MANAGER_EVENTS_QUEUE = "MANAGER_EVENT"
AMQP_MANAGER_EVENTS_ROUTING_KEY = "routingKeyManagerEvents"

AMQP_MANAGER_EVENTS_DL_X = "exchangeDLQManagerEvents"
AMQP_MANAGER_EVENTS_DL_QUEUE = "MANAGER_EVENT_DLQ"
AMQP_MANAGER_EVENTS_DL_ROUTING_KEY = "routingKeyDLQManagerEvents"

AMQP_TASKS_DL_X = "exchangeDLQTasks"
AMQP_TASKS_DL_QUEUE = "TASK_DLQ"
AMQP_TASKS_DL_ROUTING_KEY = "routingKeyDLQTasks"

AMQP_TASKS_X = "exchangeMainTasks"
AMQP_TASKS_QUEUE = "TASK"
AMQP_TASKS_ROUTING_KEY = "routingKeyMainTasks"

AMQP_WORKER_EVENTS_X = "exchangeWorkerEvents"
AMQP_WORKER_EVENTS_QUEUE = "WORKER_EVENT"
AMQP_WORKER_EVENTS_ROUTING_KEY = "routingKeyWorkerEvents"
